
import pytest

from ygoutil.card import CardType, LinkMark
from ygoutil.card.ids import CnJpEnIDCard
from ygoutil.source import BaiGe
from ygoutil.source.baige import BaiGePage, BaiGeNameUnit, BaiGeExtraUnit

@pytest.fixture(scope="module")
def baige():
    return BaiGe()

@pytest.fixture(scope="module")
def baige_page():
    return BaiGePage()

@pytest.mark.asyncio
async def test_baige(baige: BaiGe):
    card = await baige.from_id(84536654)
    assert card
    assert isinstance(card.names, BaiGeNameUnit)
    assert card.name == "形态变化" and card.names.jp_name == "フォーム・チェンジ"
    assert isinstance(card.extras, BaiGeExtraUnit)
    assert card.extras.c_id == 9068

@pytest.mark.asyncio
async def test_baige_multiple(baige: BaiGe):
    cards = await baige.list_from_query("被封印者的", limit=3)
    assert cards
    assert len(cards) == 3
    assert all("被封印者的" in card.name for card in cards)
    assert all(card.monster for card in cards)
    assert baige.current_query.query == "被封印者的"
    assert baige.current_query.current == 3

@pytest.mark.asyncio
async def test_baige_ids(baige: BaiGe):
    ids = await baige.list_ids_from_query("波塞冬", limit=10)
    assert ids
    assert len(ids) < 10
    assert all(isinstance(card, CnJpEnIDCard) for card in ids)

@pytest.mark.asyncio
async def test_baige_page(baige_page: BaiGePage):
    card = await baige_page.from_name("解码语者")
    assert card
    assert card.name == "解码语者"
    assert card.monster
    assert card.monster.link.link == 3
    assert card.monster.link.marks == LinkMark.BottomLeft | LinkMark.BottomRight | LinkMark.Top

@pytest.mark.asyncio
async def test_baige_page_multiple(baige_page: BaiGePage):
    cards = await baige_page.list_from_query("竹光", limit=6)
    assert cards
    assert len(cards) == 6
    assert all(CardType.Spell in card.types for card in cards)
    assert baige_page.current_query.query == "竹光"
    assert baige_page.current_query.current == 6

@pytest.mark.asyncio
async def test_baige_page_ids(baige_page: BaiGePage):
    ids = await baige_page.list_ids_from_query("5回合后", limit=10)
    assert ids
    assert len(ids) < 10
    assert all(isinstance(card.id, int) for card in ids)
    assert any(86100785 == card.id for card in ids)  # 区域吞噬者
    assert any(isinstance(card, CnJpEnIDCard) and "異国の剣士" == card.jp_name for card in ids)  # 异国的剑士




from typing import AsyncGenerator, TYPE_CHECKING, Literal, overload

from ygoutil.card import Card
from ygoutil.card.ids import CnJpEnIDCard
from ygoutil.card.unit import PTextUnit
from ygoutil.source.base import CardSource, QueryInfo
from ygoutil.source.baige.misc import Site, BaiGeResult, BaiGeCard, BaiGeNameUnit, BaiGeExtraUnit
from ygoutil.source.misc import get_json, USE_CLIENT_DEFAULT, deal_web_text
from ygoutil.source.cdb.misc import card_from_types, card_from_cdb_data

class BaiGe(CardSource):
    """ 查百鸽 """

    api_url = f"{Site.base_url}api/v0/"
    timeout = USE_CLIENT_DEFAULT

    def json2card(self, json: BaiGeCard) -> Card:
        data = json["data"]
        types = data["type"]
        card = card_from_types(types)
        card = card_from_cdb_data(card, json["id"], data["ot"], 0, data["setcode"], types, 
                                  data["atk"], data["def"], data["level"], data["race"], data["attribute"], 0)
        card.names = BaiGeNameUnit(card).update_from(json)
        if isinstance(card.texts, PTextUnit):
            card.texts.p_text = deal_web_text(json["text"]["pdesc"])
        card.texts._text = deal_web_text(json["text"]["desc"])
        card._extra_unit = BaiGeExtraUnit(card).update_from(json)
        return card

    def json2ids(self, json: BaiGeCard) -> CnJpEnIDCard:
        return CnJpEnIDCard(json["id"], json.get("cn_name", ""), json.get("jp_name", ""), json.get("en_name", ""))

    async def from_id(self, card_id: int) -> Card | None:
        return await self.from_query(str(card_id))
    
    async def from_name(self, card_name: str) -> Card | None:
        return await self.from_query(card_name)

    if TYPE_CHECKING:
        @overload   # 用 overload 和字面量来区分返回值
        async def _gen_from_query(self, query: str, ids_only: Literal[False] = False) -> AsyncGenerator[Card, None]:
            ...

        @overload
        async def _gen_from_query(self, query: str, ids_only: Literal[True] = True) -> AsyncGenerator[CnJpEnIDCard, None]:
            ...

    async def _gen_from_query(self, query: str, ids_only = False):
        result: BaiGeResult = await get_json(self.api_url, params=Site.query_params(query), timeout=self.timeout)
        self.current_query = QueryInfo(query, len(result["result"]))
        for card in result["result"]:
            self.current_query.current += 1
            if ids_only:
                yield self.json2ids(card)
            else:
                yield self.json2card(card)

    async def gen_from_query(self, query: str) -> AsyncGenerator[Card, None]:
        async for card in self._gen_from_query(query):
            yield card

    async def gen_ids_from_query(self, query: str) -> AsyncGenerator[CnJpEnIDCard, None]:
        async for card in self._gen_from_query(query, ids_only=True):
            yield card



# import functools
import sqlite3

from enum import Enum
from typing import Iterable, TYPE_CHECKING
from typing_extensions import Self
from pathlib import Path

from ygoutil.sqlbuilder import SQLBuilder


class ShrinkLevel(str, Enum):
    # TODO flag
    S0 = ""
    S1 = "type & 0x4000=0"
    S2 = "alias=0"
    S3 = "type & 0x48060C0=0"
    S4 = "type & 0x4802040!=0"

    No = S0
    NoToken = S1
    NoAlias = S2
    NoExtra = S3
    NoMain = S4

    @classmethod
    def fromInt(cls, i: int | Self | Iterable[int]) -> Self | str:
        if isinstance(i, cls):
            return i
        if TYPE_CHECKING:
            assert not isinstance(i, ShrinkLevel)
        if isinstance(i, Iterable):
            return " AND ".join([cls.fromInt(x) for x in i if x])  # x!=0 and x!=S0
        return getattr(cls, f"S{i}")

    # @property
    # def WHERE(self):
    #     if self:
    #         return f"WHERE {self}"
    #     return self


def WHERE(s: str):
    if s:
        return f"WHERE {s}"
    return s


class CDBReader:
    def __init__(self, path=None):
        self.cdbpath = path
        self.conn = None
        self.cursor = None

    def __enter__(self):
        self.connect()
        # print("连上啦")
        return self

    def __exit__(self, etype, eval, traceback):
        # print("关掉啦")
        self.close()

    def connect(self, path=None):
        if path:
            self.conn = sqlite3.connect(path)
        elif self.cdbpath:
            self.conn = sqlite3.connect(self.cdbpath)
        else:
            raise Exception("no cdb path")
        if self.conn:
            self.cursor = self.conn.cursor()

    def close(self):
        if self.conn:
            self.conn.commit()
            self.conn.close()
        else:
            raise Exception("no cdb connection")

    @staticmethod
    def listOfSQL(num: int):  # lens=3 -> (?,?,?)
        return "(" + ",".join(["?"] * num) + ")"

    @staticmethod
    def unique(lst):  # [111,111,222] -> [111,222]
        return list(set(lst))

    def getCardByID(self, id: int):  # id是整数
        assert self.cursor
        self.cursor.execute(
            "SELECT t.name,t.DESC,d.* FROM texts t INNER JOIN datas d ON t.id=d.id WHERE t.id=?",
            (id,),
        )
        # return self.cursor.fetchall()[0]
        return self.cursor.fetchone()

    def getCardsByIDs(self, *ids):  # ids是元祖、列表之类的
        assert self.cursor
        sql = f"SELECT t.name,t.DESC,d.* FROM texts t INNER JOIN datas d ON t.id=d.id WHERE t.id IN {CDBReader.listOfSQL(len(ids))}"
        # print(sql)
        # print(ids)
        self.cursor.execute(sql, tuple(ids))
        return self.cursor.fetchall()

    def getIDsByName(self, name):  # name是字符串 完全匹配
        assert self.cursor
        self.cursor.execute("SELECT id FROM texts WHERE name=?", (name,))
        ids = self.cursor.fetchall()
        return [x[0] for x in ids]

    def getCardsByName(self, name):
        ids = self.getIDsByName(name)
        if ids:
            return self.getCardsByIDs(*ids)

    def getCardCount(self, shrink=0):  # full 0 全部 1 无衍生物 2 无同名卡无衍生物
        assert self.cursor
        level = ShrinkLevel.fromInt(shrink)
        self.cursor.execute(f"SELECT COUNT(*) FROM datas {WHERE(level)}")
        return self.cursor.fetchall()[0]

    def getRandomIDs(self, count=1, shrink=0):
        assert self.cursor
        level = ShrinkLevel.fromInt(shrink)
        self.cursor.execute(
            f"SELECT id FROM datas {WHERE(level)} ORDER BY RANDOM() limit ?", (count,)
        )
        ids = self.cursor.fetchall()
        return [x[0] for x in ids]

    def getRandomNames(self, count=1, shrink=0):
        assert self.cursor
        ids = self.getRandomIDs(count, shrink)
        self.cursor.execute(
            f"SELECT name FROM texts WHERE id IN {CDBReader.listOfSQL(len(ids))}",
            tuple(ids),
        )
        names = self.cursor.fetchall()
        return [x[0] for x in names]

    def getIDsByInput(self, ipt, expect=()):
        pass

    def getCardsByInput(self, ipt):
        pass

    def getCardsByBuilder(self, builder: SQLBuilder, num=50):
        assert self.cursor
        tail = f"ORDER BY d.level & 15 DESC,d.type LIMIT {num}"  # 15=>0b1111
        sql = builder.resolve()
        sql = f"{sql} {tail}"
        # print(sql)
        # print(builder.params)
        if builder.params:
            self.cursor.execute(sql, tuple(builder.params))
        else:
            self.cursor.execute(sql)
        return self.cursor.fetchall()

    def getYDCards(self):
        assert self.cursor
        self.cursor.execute(
            "SELECT id FROM datas WHERE type & 0x48060C0=0 ORDER BY RANDOM() limit 60"
        )
        main = self.cursor.fetchall()
        main = [x[0] for x in main]
        self.cursor.execute(
            "SELECT id FROM datas WHERE type & 0x4802040!=0 ORDER BY RANDOM() limit 30"
        )
        extra = self.cursor.fetchall()
        extra = [x[0] for x in extra]
        return [main, extra]

class ConfReader:
    def __init__(self, path: str | Path = None):
        self.filepath = Path(path) if path else path

    def loadLFlist(self, path: str | Path = None):
        if path:
            self.filepath = Path(path)
        if TYPE_CHECKING:
            assert isinstance(self.filepath, Path)
        self.lfdict = {}
        start = False
        with open(self.filepath, "r", encoding="utf-8") as f:
            for line in f:
                if start:
                    if line.startswith("!"):
                        break
                    if line[:1].isdigit():
                        temp = line.split()
                        self.lfdict[int(temp[0])] = int(temp[1])
                else:
                    if line.startswith("!"):
                        self.lfname = line[1:]
                        start = True

    def loadSets(self, path: str | Path = None):
        if path:
            self.filepath = Path(path)
        if TYPE_CHECKING:
            assert isinstance(self.filepath, Path)
        self.setdict = {}
        with open(self.filepath, "r", encoding="utf-8") as f:
            for line in f:
                if line.startswith("!setname"):
                    temp = line.split()
                    self.setdict[int(temp[1], base=16)] = temp[2].split("\t")[0]


if __name__ == "__main__":
    conf = ConfReader()
    conf.loadLFlist("./ygo/lflist.conf")
    conf.loadSets("./ygo/strings.conf")
    # print(conf.lfname)
    # print(conf.lfdict)
    # print(conf.setdict)
    cdb = CDBReader()
    cdb.connect("./ygo/cards.cdb")
    ipt = ""
    # result=cdb.getCardsByIDs([1861629, 1861630])
    # print(result)
    """
    while ipt!="exit":
        ipt=input()
        print( cdb.getIDsByName(ipt) )
        #try:
        #    
        #except:
        #    print("不好使")
    """
    ct = cdb.getCardByID(9999)
    print(ct)
    # ids=cdb.getRandomIDs()
    # cts=cdb.getCardsByIDs(*ids)
    # c=Card()
    # c.fromCDBTuple(cts[0],conf.setdict,conf.lfdict)
    # print( c )
    cdb.close()




def check_cached_property(obj: object, name: str, default = None):
    """ 在不触发 cached_property 的情况下检查其是否已存在 """
    return obj.__dict__.get(name, default)

def line_or_not(line: str):
    return f"{line}\n" if (line := line.rstrip()) else ""





from abc import ABCMeta, abstractmethod
from typing import AsyncGenerator
from dataclasses import dataclass
from functools import cached_property

from ygoutil.card import Card
from ygoutil.card.ids import IDCard
from ygoutil.source.misc import gen_limit

@dataclass
class QueryInfo:
    """ 查询信息 """
    query: str = ""
    total: int = 0
    current: int = 0

    def __bool__(self):
        return bool(self.query or self.total or self.current)

class CardSource(metaclass=ABCMeta):
    """ 卡片源 """

    @cached_property
    def current_query(self) -> QueryInfo:
        """ 当前查询信息 """
        return QueryInfo("")

    @abstractmethod
    async def from_id(self, card_id: int) -> Card | None:
        """ 由卡号获取一张卡 """
        raise NotImplementedError

    @abstractmethod
    async def from_name(self, card_name: str) -> Card | None:
        """ 由卡名获取一张卡 """
        raise NotImplementedError

    @abstractmethod
    async def gen_from_query(self, query: str) -> AsyncGenerator[Card, None]:
        """ 由 query 获取一组卡片 """
        raise NotImplementedError
        yield

    @abstractmethod
    async def gen_ids_from_query(self, query: str) -> AsyncGenerator[IDCard, None]:
        """ 由 query 获取一组卡片标识（卡号、卡名） """
        raise NotImplementedError
        yield

    async def from_query(self, query: str):
        """ 由 query 获取一张卡 """
        return await anext(self.gen_from_query(query), None)

    async def list_from_query(self, query: str, limit: int = 100):
        """ 由 query 获取一组卡片 """
        return [card async for card in gen_limit(self.gen_from_query(query), limit)]

    async def list_ids_from_query(self, query: str, limit: int = 100):
        """ 由 query 获取一组卡片标识（卡号、卡名） """
        return [card async for card in gen_limit(self.gen_ids_from_query(query), limit)]




from ygoutil.card.card import Card
from ygoutil.card.ids import IDCard
from ygoutil.card.constant import CardType, CardRace, CardAttribute, LinkMark, CardCategory, CardOT, CardLF

__all__ = [
    'Card', 
    'IDCard',
    'CardType', 'CardRace', 'CardAttribute', 'LinkMark', 'CardCategory', 'CardOT', 'CardLF'
]




from ygoutil.source.baige.api import BaiGe
from ygoutil.source.baige.misc import BaiGeNameUnit, BaiGeExtraUnit, BaiGeURLUnit
from ygoutil.source.baige.page import BaiGePage

__all__ = ["BaiGe", "BaiGePage", "BaiGeNameUnit", "BaiGeExtraUnit", "BaiGeURLUnit"]




from pathlib import Path

import httpx
import pytest
import pytest_asyncio

from ygoutil.dataloader import CDBReader, ConfReader

class Links:
    cdb = r"https://cdn01.moecube.com/koishipro/ygopro-database/zh-CN/cards.cdb"
    strings = r"https://cdn01.moecube.com/koishipro/ygopro-database/zh-CN/strings.conf"
    lflist = r"https://cdn01.moecube.com/koishipro/ygopro-database/zh-CN/lflist.conf"

async def download(url: str, file: Path):
    async with httpx.AsyncClient() as client:
        async with client.stream("GET", url) as response:
            print(f"Downloading to {file.as_posix()}...({url})")
            with file.open("wb") as f:
                async for chunk in response.aiter_bytes():
                    f.write(chunk)
    return file

@pytest.fixture(scope="module")
def temp_dir(tmp_path_factory: pytest.TempPathFactory):
    return tmp_path_factory.mktemp("ygo")

@pytest_asyncio.fixture(scope="module")
async def cdb_file(temp_dir):
    return await download(Links.cdb, temp_dir / "cards.cdb")

@pytest_asyncio.fixture(scope="module")
async def strings_file(temp_dir: Path):
    return await download(Links.strings, temp_dir / "strings.conf")

@pytest_asyncio.fixture(scope="module")
async def lflist_file(temp_dir: Path):
    return await download(Links.lflist, temp_dir / "lflist.conf")

@pytest.fixture
def conf_reader(strings_file, lflist_file):
    reader = ConfReader()
    reader.loadLFlist(lflist_file)
    reader.loadSets(strings_file)
    return reader

@pytest.fixture
def cdb_reader(cdb_file):
    with CDBReader(cdb_file) as reader:
        yield reader
    
def test_cdb(cdb_reader: CDBReader):
    assert cdb_reader.getCardByID(10000)  # 万物创世龙
    assert cdb_reader.getCardByID(9999) is None
    assert cdb_reader.getIDsByName("解码语者")

def test_conf(conf_reader: ConfReader):
    assert conf_reader.lfname
    assert conf_reader.lfdict
    assert conf_reader.setdict



# coding:utf-8
# import sqlite3
# import configparser
import json
from typing import TYPE_CHECKING, AsyncGenerator, overload, Literal
from urllib import parse
from enum import Enum
from dataclasses import dataclass
from functools import cached_property

import httpx
from bs4 import BeautifulSoup, NavigableString, CData, Tag, ResultSet
# from bs4.element import NavigableString, CData, Tag

try:
    from bs4.element import TemplateString  # 新版bs4需要
except ImportError:
    TemplateString = NavigableString

from ygoutil.card import Card, CardAttribute, CardRace, LinkMark, CardLF
from ygoutil.card.ids import CnJpEnIDCard
from ygoutil.card.unit import CnJpEnNameUnit, RushDuelIDUnit, URLUnit, LimitUnit
from ygoutil.source.base import CardSource, QueryInfo
from ygoutil.source.misc import get_html, deal_int

class Site:
    base_url = r"https://www.ourocg.cn/"
    wiki_url = r"https://yugioh-wiki.net/"

    @classmethod
    def search_url(cls, query: str, page: int = 0):
        if page:
            return f"{cls.base_url}search/{query}/{page}"
        return f"{cls.base_url}search/{query}"

@dataclass
class OurOcgIDCard(CnJpEnIDCard):
    nw_name: str = ""

@dataclass
class OurOcgQueryInfo(QueryInfo):
    total_page: int = 0
    current_page: int = 0

class Translate(Enum):
    """ 译文偏好 """
    CN = 0
    NW = 1
    DEFAULT = 2

class OurOcg(CardSource):
    """ 查 OurOcg 网页 """
    def __init__(self):
        self.edition = Translate.DEFAULT
        self.name_search_limit = 10
        """ 
            from_name 中比对卡名时，最多比对卡片数量\n
            默认最多 10 张（即 1 页），为 0 时表示无限制
        """

    def SetTranslateEdition(self, ed: str):
        for trans in Translate:
            if trans.name == ed.upper():
                self.edition = trans
                break

    ourocgLink = Site.base_url

    def GetHTML(self, url):
        res = httpx.get(url)
        return res.text

    def GetCardHTMLWithJ(self, searchtext: str, url: str, searchHTML: str):
        targeturl: str = ""
        cardJson = None
        # Pmark=[None,None]
        html = BeautifulSoup(searchHTML, "lxml")
        if html.find_all("title")[0].string.startswith("搜索"):  # 搜索结果页
            scripts = html.find_all("script")
            for s in scripts:
                if str(s.string).strip().startswith("window.__STORE__"):
                    targets = str(s.string[s.string.find("{") :]).strip()[:-1]  # json
                    carddata = json.loads(targets)
                    sts = searchtext.strip()
                    for tar in carddata["cards"]:
                        if (
                            tar["name"] == sts
                            or tar["name_nw"] == sts
                            or tar["name_ja"] == sts
                            or tar["name_en"] == sts
                        ):
                            targeturl = tar["href"].replace("\\", "")
                            cardJson = tar
                            # Pmark=[tar.get("pend_l",None),tar.get("pend_r",None)]
                            # print(Pmark)
                    if targeturl == "" and len(carddata["cards"]) != 0:
                        tar = carddata["cards"][0]
                        targeturl = tar["href"].replace("\\", "")
                        cardJson = tar
                        # Pmark=[tar.get("pend_l",None),tar.get("pend_r",None)]
                    # print(targeturl)
                    break
        else:
            targeturl = url  # 直接是卡片详情页
        return targeturl, cardJson

    def GetCardFromHTML(self, cardHTML, cardJson) -> Card:
        if cardJson is None:
            cardJson = {}
        # print(cardJson)
        html = BeautifulSoup(cardHTML, "lxml")
        is_RD = self._is_RD(html)
        # div = html.find_all("div", {"class": "val"})
        # divr=[[y for y in x.stripped_strings] for x in div]
        divr = self._div_data(html)
        # 类似 stripped_strings，但新版本需要纳入 TemplateString 才能效果一致
        # print(divr)
        cardtypes = divr[3]
        c = Card.from_types(*cardtypes)
        if c.is_link:
            linkmark = html.find("div", {"class": "linkMark"})
            if TYPE_CHECKING:
                assert isinstance(linkmark, Tag)
            for mark in linkmark.find_all("i"):
                temp = mark["class"][1].split("_")
                if temp[2] == "on":
                    c.monster.link.marks |= LinkMark.from_number(int(temp[1]), with_mid=True)
        c.names = CnJpEnNameUnit(c)
        cn_name, nw_name, c.names.jp_name, c.names.en_name = self._names_from_div(divr)
        c.names.name = cn_name if self.edition is Translate.CN else nw_name  # 默认是 nw
        # c.isRD = isRD
        if is_RD:
            c.ids = RushDuelIDUnit(c)
        else:
            id = divr[4][0]
            id = int(id) if id and id != "-" else 0  # id 可能为 -
            c.id = id
        c._url_unit = URLUnit(c)
        c._url_unit.url = cardJson.get("href", "")    # 如果直接是卡片详情页，就没有
        c._url_unit.pic = cardJson.get("img_url", "")
        if TYPE_CHECKING:
            assert html.head and html.body
        if not c._url_unit.url:  # 试着从详情页直接获取连接
            if link_tag := html.head.find("link", {"rel": "canonical"}):
                if TYPE_CHECKING:
                    assert isinstance(link_tag, Tag)
                c._url_unit.url = str(link_tag.get("href", ""))
        if not c._url_unit.pic:
            if img_parent := html.body.find("div", {"class": "img"}):
                if TYPE_CHECKING:
                    assert isinstance(img_parent, Tag)
                if img_tag := img_parent.img:
                    c._url_unit.url = str(img_tag.get("src", ""))
        limitnum = 5
        if is_RD:
            limitnum = 4
        c._limit_unit = LimitUnit(c)
        limit = str(divr[limitnum][0])
        c._limit_unit.limit = CardLF.from_str(limit.removesuffix("卡")) or limit
        otnum = limitnum + 1
        if divr[otnum]:  # 如果是OCG/TCG专有
            c._limit_unit.ot = str(divr[otnum][0])
        effectnum = -1
        if c.is_monster:
            monster = c.monster
            monster.race = CardRace.from_str(divr[otnum + 1][0]) or CardRace.Unknown
            monster.attribute = CardAttribute.from_str(divr[otnum + 2][0]) or CardAttribute.Unknown
            # if c.isXyz:
            #     c.rank = OurOCG.dealInt(divr[otnum + 3][0])
            #     c.level = c.rank
            if monster._pmark_unit:
                # c.Pmark=Pmark
                monster._pmark_unit.mark = (cardJson.get("pend_l", None), cardJson.get("pend_r", None))
            if c.is_link:
                link = deal_int(divr[otnum + 5][0])
                attack = deal_int(divr[otnum + 4][0])
                if TYPE_CHECKING:
                    assert isinstance(link, int)
                    assert isinstance(attack, int)
                monster.link.link = link
                monster.attack = attack
            else:
                level = deal_int(divr[otnum + 3][0])
                attack = deal_int(divr[otnum + 4][0])
                defence = deal_int(divr[otnum + 5][0])
                if TYPE_CHECKING:
                    assert isinstance(level, int)
                    assert isinstance(attack, int)
                    assert isinstance(defence, int)
                monster.level = level
                monster.attack = attack
                monster.defence = defence
        L = len(divr[effectnum])
        tempString = divr[effectnum][-1]
        effectlist = [0]
        for x in range(-2, -1 * (L + 1), -1):
            if divr[effectnum][x] == tempString:
                tempnum = -1 - x
                effectlist.append(L - 2 * tempnum)
                effectlist.append(effectlist[1] + tempnum)
                effectlist.append(effectlist[2] + tempnum)
        effects = divr[effectnum][
            effectlist[self.edition.value] : effectlist[self.edition.value + 1]
        ]
        effectText = "\n".join(effects)
        c.text = OurOcg.beautifyText(effectText)
        return c

    def FindCardByName(self, searchtext):
        url = Site.search_url(searchtext)
        searchHTML = self.GetHTML(url)
        targeturl, cardJson = self.GetCardHTMLWithJ(searchtext, url, searchHTML)
        if targeturl:
            cardHTML = self.GetHTML(targeturl)
            return self.GetCardFromHTML(cardHTML, cardJson)
        return None

    # async def AsyncGetHTML(self, url):
    #     async with httpx.AsyncClient() as client:
    #         res = await client.get(url)
    #         return res.text

    # async def AsyncSearchByName(self, searchtext: str):
    #     url = Site.search_url(searchtext)
    #     searchHTML = await get_html(url)
    #     targeturl, cardJson = self.GetCardHTMLWithJ(searchtext, url, searchHTML)
    #     if targeturl:
    #         cardHTML = await get_html(targeturl)
    #         return self.GetCardFromHTML(cardHTML, cardJson)
    #     return None

    wikiLink = Site.wiki_url

    def getWikiLink(self, card: Card):
        if TYPE_CHECKING:
            assert isinstance(card.names, CnJpEnNameUnit)
        if card.names.jp_name:
            pageword = f"《{OurOcg.towikistr(card.names.jp_name)}》"
        elif card.names.en_name:
            pageword = f"《{card.names.en_name}》"
        else:
            return None
        pageword = parse.quote(pageword, encoding="euc-jp")
        return f"{self.wikiLink}index.php?cmd=read&page={pageword}"

    @staticmethod
    def towikistr(text: str):
        """半角转全角，以及一些特殊符号的转换"""
        transDict = {
            #' ':chr(12288), #半角空格直接转化
            "·": "・",
            "－": "−",
            "Ⅰ": "Ｉ",
            "Ⅱ": "ＩＩ",
            "Ⅲ": "ＩＩＩ",
            "Ⅳ": "ＩＶ",
            "Ⅴ": "Ｖ",
            "Ⅵ": "ＶＩ",
            "Ⅶ": "ＶＩＩ",
            "Ⅷ": "ＶＩＩＩ",
            "Ⅸ": "ＩＸ",
            "Ⅹ": "Ｘ",
            "Ⅺ": "ＸＩ",
            "Ⅻ": "ＸＩＩ",
        }
        r = ""
        for c in text:
            trans = transDict.get(c, None)
            if trans:
                c = trans
            else:
                oc = ord(c)
                if oc > 32 and oc <= 126:  # 半角字符（除空格）根据关系转化
                    c = chr(oc + 65248)
            r += c
        return r

    @staticmethod
    def beautifyText(text):
        """试着给效果文本加换行，好看一点"""
        nums = set(list("①②③④⑤⑥⑦⑧⑨⑩●"))
        transDict = {"・": "·"}
        r = ""
        length = len(text)
        for i, c in enumerate(text):
            trans = transDict.get(c, None)
            if trans:
                c = trans
            elif c in nums:
                if 1 < i < length - 1 and text[i - 1] != "\n" and text[i + 1] == "：":
                    c = "\n" + c
            r += c
        return r

    # ↑ 之前的老代码 ↑
    @staticmethod
    def _div_data(html: BeautifulSoup):
        div = html.find_all("div", {"class": "val"})
        # divr=[[y for y in x.stripped_strings] for x in div]
        return [
            [ y for y in x._all_strings(True, types=(NavigableString, CData, TemplateString)) ]
            for x in div
        ]

    @staticmethod
    def _is_RD(html: BeautifulSoup):
        return bool(html.find("div", {"class": "rd-mark"}))

    def _names_from_div(self, divr: list[list]):
        cn_name = divr[0][Translate.CN.value]
        nw_name = divr[0][Translate.NW.value]
        jp_name = divr[1][0]
        en_name = divr[2][0]
        if TYPE_CHECKING:
            assert isinstance(cn_name, str) and isinstance(nw_name, str) and isinstance(jp_name, str) and isinstance(en_name, str)
        jp_name = jp_name if jp_name != "-" else ""
        en_name = en_name if en_name != "-" else ""
        jp_name = jp_name.replace("・", "·")
        return cn_name, nw_name, jp_name, en_name

    def card_url_json_gen(self, url: str, html: str):
        root = BeautifulSoup(html, "lxml")
        if TYPE_CHECKING:
            assert root.head
            assert root.body
        title = root.head.find("title")
        if TYPE_CHECKING:
            assert isinstance(title, Tag)
        if not (title and title.string and title.string.startswith("搜索")):  # 不是搜索结果页，直接是卡片详情页
            yield url, None                                                   # 或者是卡名以“搜索”开头的卡（
            return
        scripts: ResultSet[Tag] = root.body.find_all("script")
        script_text = ""
        # 试着拿到 json 所在的脚本
        if not any(script_text := s.string
                   for s in scripts 
                   if s.string and s.string.strip().startswith("window.__STORE__")):
            return
        card_jsons = script_text[script_text.find("{"):].strip()[:-1]  # 去掉 { 之前的部分和尾部分号
        card_data: dict[str, dict] = json.loads(card_jsons)
        if meta := card_data.get("meta"):   # {keyword, count, total_page, cur_page}  试着更新查询信息
            self.current_query.query = meta.get("keyword", self.current_query.query)  # 也更新 query
            self.current_query.total = meta.get("count", self.current_query.total)
            self.current_query.total_page = meta.get("total_page", self.current_query.total_page)
            self.current_query.current_page = meta.get("cur_page", self.current_query.current_page)
        for one in card_data["cards"]:
            if TYPE_CHECKING:
                assert isinstance(one, dict)
            target_url = str(one["href"]).replace("\\", "")
            yield target_url, one

    def ids_from_json(self, card_json: dict | None) -> OurOcgIDCard | None:
        if card_json:
            id = int(card_json.get("password") or 0)
            cn_name = card_json.get("name") or ""
            nw_name = card_json.get("name_nw") or ""
            jp_name = card_json.get("name_ja") or ""
            en_name = card_json.get("name_en") or ""  # 名称可能为 None
            return OurOcgIDCard(id, cn_name, jp_name, en_name, nw_name)

    def ids_from_html(self, card_html) -> OurOcgIDCard:
        html = BeautifulSoup(card_html, "lxml")
        is_RD = self._is_RD(html)
        divr = self._div_data(html)
        cn_name, nw_name, jp_name, en_name = self._names_from_div(divr)
        id = 0 if is_RD else int(divr[4][0])
        return OurOcgIDCard(id, cn_name, jp_name, en_name, nw_name)

    @cached_property
    def current_query(self):
        return OurOcgQueryInfo()

    if TYPE_CHECKING:
        @overload   # 用 overload 和字面量来区分返回值
        async def search_gen(self, query: str, ids_only: Literal[False] = False) -> AsyncGenerator[Card, None]:
            ...

        @overload
        async def search_gen(self, query: str, ids_only: Literal[True] = True) -> AsyncGenerator[OurOcgIDCard, None]:
            ...

    async def search_gen(self, query: str, ids_only = False):
        self.current_query = query_info = OurOcgQueryInfo(query)
        while query_info.current_page <= query_info.total_page:
            url = Site.search_url(query, query_info.current_page)  # 默认最初 current_page = 0
            html = await get_html(url, follow_redirects=True)      # OurOcg 有页面重定向
            for card_url, card_json in self.card_url_json_gen(url, html):
                if ids_only and (ids := self.ids_from_json(card_json)):
                    query_info.current += 1
                    yield ids
                    continue
                if card_url:
                    card_html = await get_html(card_url, follow_redirects=True)
                    query_info.current += 1
                    if ids_only:
                        yield self.ids_from_html(card_html)
                    else:
                        yield self.GetCardFromHTML(card_html, card_json)
            query_info.current_page += 1

    async def from_id(self, card_id: int) -> Card | None:
        return await self.from_query(str(card_id))

    async def from_name(self, card_name: str) -> Card | None:
        result = None
        async for card in self.search_gen(card_name, ids_only=True):
            # print(card)
            # print(self.current_query)
            if result is None:
                result = card  # 第一张卡
            if card_name in (card.name, card.nw_name, card.jp_name, card.en_name):
                result = card
                break
            if self.name_search_limit and self.current_query.current >= self.name_search_limit:    
                # 超过比对上限而没找到卡名完全匹配的，返回第一张卡
                break
        if result:
            # 相当于第二次 search_gen，会覆盖 current_query
            return await self.from_query(str(result.id or result.name or result.jp_name or result.en_name))
    
    AsyncSearchByName = from_name  # 兼容

    async def gen_from_query(self, query: str) -> AsyncGenerator[Card, None]:
        async for card in self.search_gen(query):
            yield card

    async def gen_ids_from_query(self, query: str) -> AsyncGenerator[OurOcgIDCard, None]:
        async for card in self.search_gen(query, ids_only=True):
            yield card

if __name__ == "__main__":
    # text=input()
    # a=ourocg()
    # print(a.FindCardByName(text))

    async def main():
        text = input()
        a = OurOcg()
        print(await a.AsyncSearchByName(text))

    import asyncio

    asyncio.run(main())




from functools import cached_property

from ygoutil.card.constant import CardType, CardCategory
from ygoutil.card.unit import (IDUnit, NameUnit, TextUnit, LimitUnit, 
                               SetUnit, TypeUnit, CategoryUnit, MonsterUnit, URLUnit, CardUnit)
from ygoutil.card.misc import line_or_not

class Card:
    """ YGO 卡片 """

    @cached_property
    def ids(self):
        return IDUnit(self)

    @cached_property
    def names(self):
        return NameUnit(self)

    @cached_property
    def texts(self):
        return TextUnit(self)

    @cached_property
    def _limit_unit(self) -> LimitUnit | None:
        return None

    @cached_property
    def _set_unit(self) -> SetUnit | None:
        return None
    
    @cached_property
    def _type_unit(self):
        return TypeUnit(self)
    
    @cached_property
    def _category_unit(self) -> CategoryUnit | None:
        return None
    
    @cached_property
    def _monster_unit(self) -> MonsterUnit | None:
        return None
    
    @cached_property
    def _url_unit(self) -> URLUnit | None:
        return None

    @cached_property
    def _extra_unit(self) -> CardUnit | None:
        return None

    @property
    def id(self):
        return self.ids.id
    
    @id.setter
    def id(self, value: int | str):
        self.ids.id = value

    @property
    def name(self):
        return self.names.name
    
    @name.setter
    def name(self, value: str):
        self.names.name = value
    
    @property
    def text(self):
        return self.texts.text
    
    @text.setter
    def text(self, value: str):
        self.texts.text = value

    @property
    def limits(self):
        if self._limit_unit:
            return self._limit_unit
        raise AttributeError("Card does not have limitation data")

    @property
    def sets(self):
        if self._set_unit:
            return self._set_unit.sets
        raise AttributeError("Card does not have sets")

    @sets.setter
    def sets(self, value: set[str]):
        if not self._set_unit:
            self._set_unit = SetUnit(self)
        self._set_unit.sets = value

    @property
    def types(self):
        return self._type_unit.types

    @types.setter
    def types(self, value: set[CardType | str]):
        self._type_unit.types = value

    @property
    def categories(self):
        if self._category_unit:
            return self._category_unit.categories
        raise AttributeError("Card does not have categories")

    @categories.setter
    def categories(self, value: CardCategory):
        if not self._category_unit:
            self._category_unit = CategoryUnit(self)
        self._category_unit.categories = value

    @property
    def monster(self):
        if self._monster_unit:
            return self._monster_unit
        raise AttributeError("Card is not a monster")

    @property
    def urls(self):
        if self._url_unit:
            return self._url_unit
        raise AttributeError("Card does not have urls")

    @property
    def extras(self):
        if self._extra_unit:
            return self._extra_unit
        raise AttributeError("Card does not have extra data")

    @property
    def is_monster(self):
        """ 是怪兽 """
        return self._type_unit.is_monster

    @property
    def is_effect(self):
        """ 有效果 """
        return self._type_unit.is_effect

    @property
    def is_xyz(self):
        """ 是超量 """ 
        return self._type_unit.is_xyz

    @property
    def is_pendulum(self):
        """ 是灵摆 """ 
        return self._type_unit.is_pendulum

    @property
    def is_link(self):
        """ 是连接 """ 
        return self._type_unit.is_link

    @classmethod
    def from_types(cls, *types: CardType | str):
        """ 从 types 创建卡片 """
        card = Card()
        card._fill_types(*types)
        card._init_monster()
        return card

    def _fill_types(self, *types: CardType | str):
        for t in types:
            if isinstance(t, str) and (ct := CardType.from_str(t)):
                self.types.add(ct)
            else:
                self.types.add(t)

    def _init_monster(self):
        if self.is_monster:
            self._monster_unit = MonsterUnit(self)

    def _info_gen(self):
        yield line_or_not(self.names.info())             # 卡名
        yield line_or_not(self._type_unit.info())        # 类型
        line = f"{self.ids.info()}"                      # 卡号、禁限、OT
        if self._limit_unit:
            line = f"{line}  {self._limit_unit.info()}"
        yield line_or_not(line)
        if self._set_unit:
            yield line_or_not(self._set_unit.info())     # 字段
        if self._monster_unit:
            yield line_or_not(self._monster_unit.info()) # 怪兽
        yield self.texts.info()                          # 卡片文本，不用再换行
        
    def info(self):
        """ 卡片信息 """
        return "".join(self._info_gen())
        
    def __str__(self):
        return self.info()

    def __repr__(self):
        return f"{self.__class__.__name__}({self.id!r}, {self.name!r}, {self.types!r})"



from __future__ import annotations

from typing import TYPE_CHECKING

from functools import cached_property

if TYPE_CHECKING:
    from typing import Literal
    from ygoutil.card.card import Card

from ygoutil.card.constant import (CardType, CardCategory, CardRace, CardAttribute, LinkMark, LinkMarkStyle, 
                                   CardOT, CardLF)
from ygoutil.card.misc import check_cached_property, line_or_not

class CardUnit:
    """ 卡片组成部分 """

    def __init__(self, owner: Card):
        self._owner = owner

    def info(self) -> str:
        return ""

class IDUnit(CardUnit):
    """ 卡号 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.id: int | str = 0

    def info(self):
        return str(self.id)

class AliasIDUnit(IDUnit):
    """ 同名卡卡号 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        # self.id: int | str = 0
        self.alias_id: int | str = 0

class RushDuelIDUnit(IDUnit):
    """ RD 卡无卡号 """

    def __init__(self, owner: Card):
        super(IDUnit, self).__init__(owner)
        self.id = "RUSH DUEL"

class NameUnit(CardUnit):
    """ 卡名 """
    
    def __init__(self, owner: Card):
        super().__init__(owner)
        self.name: str = ""

    def info(self):
        return self.name

class CnJpEnNameUnit(NameUnit):
    """ 中日英卡名 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.jp_name: str = ""
        self.jp_ruby: str = ""
        self.en_name: str = ""

    @property
    def cn_name(self):
        return self.name
    
    @cn_name.setter
    def cn_name(self, value: str):
        self.name = value

    def _jp_info(self):
        if self.jp_ruby and self.jp_ruby != self.jp_name:
            return f"{self.jp_name}（{self.jp_ruby}）"
        return self.jp_name

    def info(self):
        return f"{line_or_not(super().info())}{line_or_not(self._jp_info())}{line_or_not(self.en_name)}".rstrip()


class TextUnit(CardUnit):
    """ 卡片文本 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self._text: str = ""
    
    @property
    def text(self):
        """ 文本 """
        return self._text
    
    @text.setter
    def text(self, value: str):
        self._text = value

    def info(self):
        return self.text

class PTextUnit(TextUnit):
    """ 灵摆效果文本 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.p_text: str = ""

    @property
    def splitter(self):
        return "【怪兽效果】" if self._owner.is_effect else "【怪兽描述】"

    @property
    def text(self):
        return f"{self.p_text.rstrip()}\n{self.splitter}\n{self._text}"

    @text.setter
    def text(self, value: str):
        self.p_text, self._text = value.split(self.splitter, maxsplit=1) # 切分灵摆效果和怪兽效果
        self.p_text = self.p_text.rstrip()
        if self.p_text.startswith("←") and (idx := self.p_text.find("→")) > 0:
            self.p_text = self.p_text[idx + 1:].lstrip()    # 试着除去灵摆效果开头的 ← ...【灵摆】... →
        self._text = self._text.lstrip()

    @classmethod
    def from_base(cls, base_unit: TextUnit):
        unit = cls(base_unit._owner)
        unit._text = base_unit._text
        return unit

class LimitUnit(CardUnit):
    """ 卡片可用性 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.ot: str | CardOT = CardOT.Unknown
        self.limit: str | CardLF = CardLF.无限制

    def _ot_info(self):
        if isinstance(self.ot, CardOT):
            return { CardOT.OCG: "OCG", CardOT.TCG: "TCG", CardOT.Custom: "DIY", CardOT.SC: "简中" }.get(self.ot, "")
        return self.ot

    def _limit_info(self):
        if self.limit is CardLF.无限制:  # 无限制不显示
            return ""
        return str(self.limit)

    def info(self):
        if self.ot:
            return f"{self._limit_info()}  {self._ot_info()}"
        return self._limit_info()

class SetUnit(CardUnit):
    """ 卡片系列（字段） """

    @cached_property
    def sets(self) -> set[str]:
        return set()

    def info(self):
        return " ".join(self.sets)

class TypeUnit(CardUnit):
    """ 卡片类型 """

    @cached_property
    def types(self) -> set[CardType | str]:  # set[str] 兼容一些未定义的类型
        return set()

    @property
    def is_monster(self):
        """ 是怪兽 """
        return CardType.Monster in self.types

    @property
    def is_effect(self):
        """ 有效果 """
        return CardType.Effect in self.types

    @property
    def is_xyz(self):
        """ 是超量 """ 
        return CardType.Xyz in self.types

    @property
    def is_pendulum(self):
        """ 是灵摆 """ 
        return CardType.Pendulum in self.types

    @property
    def is_link(self):
        """ 是连接 """ 
        return CardType.Link in self.types
    
    def info(self):
        return " ".join(str(ct) for ct in self.types)

class CategoryUnit(CardUnit):
    """ 效果分类 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.categories: CardCategory = CardCategory.未知

class MonsterUnit(CardUnit):
    """ 怪兽卡 """

    @cached_property
    def levels(self):
        return LevelUnit(self._owner)
    
    @cached_property
    def _pmark_unit(self) -> PMarkUnit | None:
        return None

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.attribute: CardAttribute = CardAttribute.Unknown
        self.race: CardRace = CardRace.Unknown
        self.attack: int | Literal["?"] | None = None
        self.defence: int | Literal["?"] | None = None
        self._init_units()

    def _init_units(self):
        if self._owner.is_xyz:
            self.levels = RankUnit(self._owner)
        if self._owner.is_pendulum:
            self._pmark_unit = PMarkUnit(self._owner)
        if self._owner.is_link:
            self.levels = LinkUnit(self._owner)

    @property
    def level(self):
        return self.levels.level

    @level.setter
    def level(self, value: int):
        self.levels.level = value

    @property
    def xyz(self):
        if isinstance(self.levels, RankUnit):
            return self.levels
        raise AttributeError("Not an xyz monster")

    @property
    def pendulum(self):
        if self._pmark_unit:
            return self._pmark_unit
        raise AttributeError("Not a pendulum monster")

    @property
    def link(self):
        if isinstance(self.levels, LinkUnit):
            return self.levels
        raise AttributeError("Not a link monster")

    def _atk_def_info(self):
        attack = self.attack or "-"
        defense = self.defence  or "-"
        return f"{attack} / {defense}" if attack and defense else ""

    def _info_gen(self):
        line = ""
        if self.race:
            line = f"{self.race}族"
        if self.attribute:
            line = f"{line}  {self.attribute}属性"
        if level_info := self.levels.info():
            line = f"{line}  {level_info}"
        yield line_or_not(line)
        yield line_or_not(self._atk_def_info())
        if isinstance(self.levels, LinkUnit):
            yield line_or_not(self.levels._mark_info())  # 连接标识
        if self._pmark_unit:
            yield line_or_not(self._pmark_unit.info())

    def info(self):
        return "".join(self._info_gen())

class LevelUnit(CardUnit):
    """ 怪兽等级 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self._level: int | None = None
    
    @property
    def level(self):
        """ 怪兽等级 """
        return self._level
    
    @level.setter
    def level(self, value: int):
        self._level = value

    def info(self):
        return f"{self._level}星" if self._level else ""

class RankUnit(LevelUnit):
    """ 超量怪兽阶级 """

    def __init__(self, owner: Card):
        super(LevelUnit, self).__init__(owner)
        self.rank: int | None = None

    @property
    def level(self):
        """ 超量怪兽阶级 """
        return self.rank
    
    @level.setter
    def level(self, value: int):
        self.rank = value

    def info(self):
        return f"{self.rank}阶" if self.rank else ""
    
class LinkUnit(LevelUnit):
    """ 链接怪兽 LINK 数 """

    def __init__(self, owner: Card):
        super(LevelUnit, self).__init__(owner)
        self.link: int | None = None
        self.marks: LinkMark = LinkMark.Unknown

    @property
    def level(self):
        """ 链接怪兽 LINK 数 """
        return self.link
    
    @level.setter
    def level(self, value: int):
        self.link = value

    def info(self):
        return f"LINK-{self.link}" if self.link else ""
    
    def _mark_info(self):
        empty = LinkMarkStyle.middle()
        marks_str = [
            str(mark) if mark in self.marks else empty 
            for mark in LinkMark if mark  # ignore LinkMark(0x00)
        ]
        lines: list[tuple[str, str, str]] = [
            (marks_str[5], marks_str[6], marks_str[7]),
            (marks_str[3], empty       , marks_str[4]),
            (marks_str[0], marks_str[1], marks_str[2])
        ]
        return "\n".join(  # join all line
            ''.join(line) for line in lines  # join one line
            if any(mark is not empty for mark in line)  # if line is not empty
        )


class PMarkUnit(CardUnit):
    """ 灵摆刻度 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.left: int | None = None
        self.right: int | None = None
        # self._init_p_text()  # 默认不用 PTextUnit，只有当效果文本能区分怪效和灵摆效果时才用

    @property
    def mark(self):
        """ 灵摆刻度 """
        return self.left, self.right

    @mark.setter
    def mark(self, value: tuple[int, int]):
        self.left, self.right = value

    def _init_p_text(self):
        if base_unit := check_cached_property(self._owner, "texts"):
            self._owner.texts = PTextUnit.from_base(base_unit)
        else:
            self._owner.texts = PTextUnit(self._owner)

    def info(self):
        return f"←{self.left} 【灵摆】 {self.right}→" if self.left and self.right else ""

class URLUnit(CardUnit):
    """ 相关链接 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.url: str = ""
        """ 卡片数据来源链接 """
        self.pic: str = ""
        """ 卡图链接 """




from typing import TypedDict
from typing_extensions import NotRequired

from ygoutil.card.card import Card
from ygoutil.card.unit import CnJpEnNameUnit, CardUnit, URLUnit

class Site:
    base_url = "https://ygocdb.com/"
    pics_url = "https://cdn.233.momobako.com/ygopro/pics/"

    @classmethod
    def card_pic_url(cls, card_id: int | str):
        return f"{cls.pics_url}{card_id}.jpg"

    @classmethod
    def card_page_url(cls, card_id: int | str):
        return f"{cls.base_url}card/{card_id}"
    
    @classmethod
    def faq_url(cls, faq_id: int):
        return f"{cls.base_url}faq/{faq_id}"

    @staticmethod
    def query_params(query: str):
        return {"search": query}

class CardText(TypedDict):
    types: str  # 例：[怪兽|效果] 种族/属性\n[★3] 1300/400
    pdesc: str
    desc: str

CardData = TypedDict("CardData", {
    "ot": int, 
    "setcode": int, 
    "type": int,
    "atk": int,
    "def": int,     # def 是关键字，所以 CardData 只能用这种形式定义
    "level": int,
    "race": int,
    "attribute": int
})

class CardHTML(TypedDict):
    pdesc: str
    desc: str
    refer: dict[str, bool]

class BaiGeCard(TypedDict):
    cid: int
    id: int

    cn_name: NotRequired[str]
    sc_name: NotRequired[str]
    md_name: NotRequired[str]
    nwbbs_n: NotRequired[str]
    cnocg_n: NotRequired[str]
    jp_ruby: NotRequired[str]
    jp_name: NotRequired[str]
    en_name: NotRequired[str]

    text: CardText
    data: CardData
    html: CardHTML

    weight: int
    faqs: list[str]
    artid: int

class BaiGeResult(TypedDict):
    result: list[BaiGeCard]
    next: int

class BaiGeNameUnit(CnJpEnNameUnit):
    """ 百鸽卡名 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.sc_name: str = ""
        self.md_name: str = ""
        self.nwbbs_name: str = ""
        self.cnocg_name: str = ""

    def update_from(self, card: BaiGeCard):
        self.cn_name = card.get("cn_name", "")
        self.sc_name = card.get("sc_name", "")
        self.md_name = card.get("md_name", "")
        self.nwbbs_name = card.get("nwbbs_n", "")
        self.cnocg_name = card.get("cnocg_n", "")
        self.jp_name = card.get("jp_name", "")
        self.jp_ruby = card.get("jp_ruby", "")
        self.en_name = card.get("en_name", "")
        return self

class BaiGeExtraUnit(CardUnit):
    """ 百鸽额外信息 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.c_id: int = 0                    # 百鸽数据库编号
        self.weight: int = 0
        self.art_id: int = 0
        self.refers: list[int] | None = None  # 效果中关联到的卡的卡号
        self.faqs: list[int] | None = None    # FAQ 号

    def update_from(self, card: BaiGeCard):
        self.c_id = card["cid"]
        self.weight = card["weight"]
        self.art_id = card["artid"]
        self.refers = [int(id) for id in card["html"]["refer"]]
        self.faqs = [int(fid) for fid in card["faqs"]]
        return self

class BaiGeURLUnit(URLUnit):
    """ 百鸽链接 """

    def __init__(self, owner: Card):
        super().__init__(owner)
        self.database_jp: str = ""
        self.database_en: str = ""
        self.database_cn: str = ""
        self.QA: str = ""
        self.wiki: str = ""
        self.yugipedia: str = ""
        self.ourocg: str = ""
        self.script: str = ""
        self.ocg_rule: str = ""

    @property
    def database(self):
        return self.database_jp
    
    @database.setter
    def database(self, value: str):
        self.database_jp = value



# import json

from typing import TYPE_CHECKING, AsyncGenerator, Generator, Literal, overload
# from urllib import parse
import httpx
from bs4 import BeautifulSoup, Tag, ResultSet
# from bs4.element import PageElement, Tag
# try:
#     from bs4.element import TemplateString #新版bs4需要
# except:
#     TemplateString=NavigableString

from ygoutil.card import Card, CardAttribute, CardRace, LinkMark
from ygoutil.card.ids import CnJpEnIDCard
from ygoutil.card.unit import LinkUnit, PTextUnit  #, RankUnit
from ygoutil.source.base import CardSource, QueryInfo
from ygoutil.source.baige.misc import Site, BaiGeNameUnit, BaiGeURLUnit, BaiGeExtraUnit
from ygoutil.source.misc import get_html, deal_int


class BaiGePage(CardSource):
    """ 查百鸽网页 """

    @property
    def link(self):
        return Site.base_url

    link_mark = {
        "↙": LinkMark.BottomLeft,
        "↓": LinkMark.Bottom,
        "↘": LinkMark.BottomRight,
        "←": LinkMark.Left,
        "→": LinkMark.Right,
        "↖": LinkMark.TopLeft,
        "↑": LinkMark.Top,
        "↗": LinkMark.TopRight,
    }

    link_name = {"数据库": "database", "db": "database", 
                "日文": "database_jp", "英文": "database_en", "简中": "database_cn",
                "q&a": "QA", "脚本": "script", "裁定": "ocg_rule"}

    @staticmethod
    def img_link(card_id):
        return Site.card_page_url(card_id)

    def get_html(self, url):
        res = httpx.get(url, timeout=20)
        return res.text

    # async def asyncGetHTML(self, url):
    #     async with httpx.AsyncClient(timeout=20) as client:
    #         res = await client.get(url)
    #         return res.text

    def html2Card(self, html, card_page=False) -> Generator[Card, None, None]:
        html = BeautifulSoup(html, "lxml")
        divs = self._html2divs(html, card_page)
        if not divs:
            return None
        # cards = []
        self.current_query.total = len(divs)
        for div in divs:
            div: Tag
            info = self._div_card_info(div)
            if card_page and len(info) == 2:  # 详情页少一个 div
                # info[2] = info[2].parent
                info.append(info[-1])
            types = None
            data = None
            desc = ""
            pdesc = ""
            state = "type"
            if TYPE_CHECKING:
                assert info[2].div
            for child in info[2].div.children:  # 效果文本
                if state == "type":
                    if isinstance(child, str):
                        child = child.strip()
                        if child:
                            types, *data = child.split()
                            types = types[1:-1].split("|")  # [魔法|永续]
                            state = "data"
                elif state == "data":
                    if isinstance(child, Tag):
                        if child.name == "hr":
                            state = "desc"
                    elif isinstance(child, str):
                        child = child.strip()
                        if child:
                            data = (data or []) + child.split()
                elif state == "desc":   # 描述文本
                    if isinstance(child, Tag):
                        if child.name == "br":
                            desc += "\n"
                        elif child.name == "hr" and types and "灵摆" in types:
                            pdesc = f"{desc.rstrip()}\n"    # 区分灵摆效果和怪兽效果
                            # desc += "\n{p}"
                            desc = ""
                        else:
                            child = child.text
                    if isinstance(child, str):
                        desc += child
            if not types:
                return None
            
            c = Card.from_types(*types)
            if TYPE_CHECKING:
                assert info[1].h2
            names = c.names = BaiGeNameUnit(c)
            c._extra_unit = BaiGeExtraUnit(c)
            (c.id, c._extra_unit.c_id, names.cn_name, names.jp_name, names.en_name, names.jp_ruby) = self._div_id_names(info)
            # c.database,c.QA,c.wiki,c.yugipedia,c.ygorg,c.ourocg,c.script,c.ocgRule=[
            #     a.get("href") for a in info[1].find_all("a")]
            # links={ a.text.strip().lower() : a.get("href") for a in info[1].find_all("a")}
            c._url_unit = BaiGeURLUnit(c)
            for atag in info[1].find_all("a"):
                atag: Tag
                aname: str = atag.text.strip().lower()  # 链接对应的名称都变成小写了
                alink = atag.get("href")
                if TYPE_CHECKING:
                    assert isinstance(alink, str)
                if aname in self.link_name:
                    setattr(c._url_unit, self.link_name[aname], alink)
                elif hasattr(c, aname):
                    setattr(c._url_unit, aname, alink)
                elif aname.startswith("详情"):
                    if alink.endswith("#faq"):  # /card/xxxx#faq
                        alink = alink[:-4]
                    c.urls.url = f"{self.link[:-1]}{alink}"  # https://ygocdb.com/card/xxxx

            # c.database=links.get("数据库")
            # c.QA=links.get("q&a")
            # c.wiki=links.get("wiki")
            # c.yugipedia=links.get("yugipedia")
            # # c.ygorg=links.get("ygorg") # 好像不再提供这个链接了
            # c.ourocg=links.get("ourocg")
            # c.script=links.get("脚本")
            # c.ocgRule=links.get("裁定")
            # print(links)
            if TYPE_CHECKING:
                assert info[0].img
            img = info[0].img.get("data-original")  # 卡图链接
            if TYPE_CHECKING:
                assert isinstance(img, str)
            c._url_unit.pic = img
            if isinstance(img, str) and img.endswith("!half"):
                c._url_unit.pic = img[:-5]
            if c.is_monster:
                if TYPE_CHECKING:
                    assert data
                race, attribute = data.pop(0).split("/")  # 恐龙/地
                race = CardRace.from_str(race)
                attribute = CardAttribute.from_str(attribute)
                if TYPE_CHECKING:
                    assert race
                    assert attribute
                c.monster.race = race
                c.monster.attribute = attribute
                lv = ""
                for char in data.pop(0):  # [★2] [☆4] [LINK-3]
                    if char.isdigit():
                        lv += char
                lv = deal_int(lv)
                if TYPE_CHECKING:
                    assert isinstance(lv, int)
                c.monster.level = lv
                attack, defence = data.pop(0).split("/")  # 1400/1200
                attack = deal_int(attack)
                defence = deal_int(defence)
                if TYPE_CHECKING:
                    assert isinstance(attack, int)
                    assert isinstance(defence, int)
                c.monster.attack = attack
                c.monster.defence = defence
                # if isinstance(c.monster.levels, RankUnit):
                #     c.monster.levels.rank = lv
                if isinstance(c.monster.levels, LinkUnit):
                    # c.monster.levels.link = lv
                    c.monster.defence = None
                    marks = data.pop(0).split("]")  # [↑][←][→][↙][↓][↘]
                    for m in marks:
                        m = m.strip("[")
                        if m:
                            c.monster.levels.marks |= self.link_mark[m]
                if c.monster._pmark_unit:
                    if data:
                        marks = data.pop(0).split("/")  # 4/4
                    else:
                        # 刻度是 0/0 时好像不会展示在网页上
                        marks = ("0", "0")
                    marks = tuple(deal_int(m) for m in marks)
                    if TYPE_CHECKING:
                        marks = int(marks[0]), int(marks[1])
                    c.monster._pmark_unit.mark = marks
                    # pdesc = f"←{marks[0]} 【灵摆】 {marks[1]}→" + pdesc
                    # if c.is_effect:
                    #     desc = desc.format(p="【怪兽效果】")
                    # else:
                    #     desc = desc.format(p="【怪兽描述】")
                    c.texts = PTextUnit(c)
                    c.texts.p_text = pdesc.strip()
            c.texts._text = desc.strip()
            # cards.append(c)
            yield c

            # return c  # 先只找一张卡

    def search(self, text: str | int, by_id=False):
        if by_id:
            url = f"{self.link}card/{text}"  # text是卡号
        else:
            url = f"{self.link}?search={text}"
        html = self.get_html(url)
        return next(self.html2Card(html, card_page=by_id), None)

    if TYPE_CHECKING:
        @overload
        async def async_search_gen(self, text: str | int, by_id=False, ids_only : Literal[False] = False) -> AsyncGenerator[Card, None]:
            ...

        @overload
        async def async_search_gen(self, text: str | int, by_id=False, ids_only : Literal[True] = True) -> AsyncGenerator[CnJpEnIDCard, None]:
            ...

    async def async_search_gen(self, text: str | int, by_id=False, ids_only=False):
        text = str(text)
        if by_id:
            url = Site.card_page_url(text)  # text是卡号
            params = None
        else:
            url = f"{self.link}"
            params = Site.query_params(text)
        html = await get_html(url, params=params, timeout=20)
        self.current_query = QueryInfo(text)   # 在这里初始化 QueryInfo，这样能覆盖所有 CardSource 要求的方法
        gen = self.html2Card(html, card_page=by_id) if not ids_only else self.html2IDCard(html)
        for card in gen:
            self.current_query.current += 1
            yield card

    async def async_search(self, text: str | int, by_id=False):
        return await anext(self.async_search_gen(text, by_id), None)

    asyncSearch = async_search

    # ↑ 之前的老代码 ↑

    @staticmethod
    def _html2divs(html: BeautifulSoup, card_page=False) -> ResultSet[Tag]:
        if card_page:
            return html.find_all("div", {"class": "row card detail"})
        else:
            return html.find_all("div", {"class": "row card result"})

    @staticmethod
    def _div_card_info(div: Tag) -> ResultSet[Tag]:
        return div.find_all("div", limit=3, recursive=False)

    @staticmethod
    def _div_id_names(info: ResultSet[Tag]):
        if TYPE_CHECKING:
            assert info[1].h2
        cn_name = info[1].h2.text  # 卡名
        jp_name = jp_ruby = en_name = ""
        extras: ResultSet[Tag] = info[1].find_all("h3")
        id_tags: ResultSet[Tag] = extras.pop().find_all("span")
        id = int(id_tags[0].text)  # 卡号
        c_id = int(id_tags[-1].text)  # 数据库编号
        # en jp 卡名
        if len(extras) < 2:  # 缺 jp 或 en
            if extras[-1].text.isascii():
                en_name = extras.pop().text
            else:
                jp_name = extras.pop().text
        else:
            en_name = extras.pop().text
            jp_name = extras.pop().text
        if info[1].h4:
            jp_ruby = info[1].h4.text
        return id, c_id, cn_name, jp_name, en_name, jp_ruby

    async def from_id(self, card_id: int):
        return await self.async_search(card_id, by_id=True)
    
    async def from_name(self, card_name: str):
        return await self.async_search(card_name)

    async def gen_from_query(self, query: str) -> AsyncGenerator[Card, None]:
        async for card in self.async_search_gen(query):
            yield card

    def html2IDCard(self, html) -> Generator[CnJpEnIDCard, None, None]:
        html = BeautifulSoup(html, "lxml")
        divs = self._html2divs(html)
        if not divs:
            return None
        self.current_query.total = len(divs)
        for div in divs:
            info = self._div_card_info(div)
            id, c_id, cn_name, jp_name, en_name, jp_ruby = self._div_id_names(info)
            yield CnJpEnIDCard(id, cn_name, jp_name, en_name)

    async def gen_ids_from_query(self, query: str) -> AsyncGenerator[CnJpEnIDCard, None]:
        async for card in self.async_search_gen(query, ids_only=True):
            yield card

if __name__ == "__main__":

    async def main():
        text = input()
        a = BaiGePage()
        print(await a.async_search(text))

    import asyncio

    asyncio.run(main())



from __future__ import annotations

from typing import AsyncGenerator, TypeVar, TYPE_CHECKING

from httpx import AsyncClient, USE_CLIENT_DEFAULT
if TYPE_CHECKING:
    from httpx._types import TimeoutTypes, QueryParamTypes
    from httpx._client import UseClientDefault

YieldT = TypeVar("YieldT")
SendT = TypeVar("SendT")

async def gen_limit(gen: AsyncGenerator[YieldT, SendT], limit: int) -> AsyncGenerator[YieldT, SendT]:
    """ 限制生成器的长度 """
    count = 0
    if count >= limit:
        return
    async for item in gen:
        yield item
        count += 1
        if count >= limit:
            break

async def get_json(url: str, params: QueryParamTypes = None, timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT, *args, **kw):
    async with AsyncClient() as client:
        response = await client.get(url, params=params, timeout=timeout, *args, **kw)
        return response.json()
    
async def get_html(url: str, params: QueryParamTypes = None, timeout: TimeoutTypes | UseClientDefault = USE_CLIENT_DEFAULT, *args, **kw):
    async with AsyncClient() as client:
        response = await client.get(url, params=params, timeout=timeout, *args, **kw)
        return response.text

def deal_int(text: str):
    return int(text) if text.isdigit() else text

def deal_web_text(text: str):
    return text.replace("\r\n", '\n').strip()




import pytest

from ygoutil.card import CardType
from ygoutil.source import OurOcg
from ygoutil.source.ourocg import CnJpEnNameUnit, RushDuelIDUnit, Translate, OurOcgIDCard

@pytest.fixture
def ourocg():
    return OurOcg()

@pytest.mark.asyncio
async def test_ourocg(ourocg: OurOcg):
    card = await ourocg.from_id(23512906)
    assert card
    assert isinstance(card.names, CnJpEnNameUnit)
    assert card.names.en_name == "Gold Pride - Leon"
    assert ourocg.current_query.current == 1

@pytest.mark.asyncio
async def test_ourocg_rd(ourocg: OurOcg):
    card = await ourocg.from_query("七道魔导士")
    assert card
    assert isinstance(card.ids, RushDuelIDUnit)
    assert card.id == "RUSH DUEL"
    assert ourocg.current_query.total_page > 1

@pytest.mark.asyncio
async def test_ourocg_from_name(ourocg: OurOcg):
    ourocg.name_search_limit = 0
    ourocg.edition = Translate.CN
    card = await ourocg.from_name("元素英雄 新宇侠")
    assert card
    assert card.name == "E·HERO 新生人"

@pytest.mark.asyncio
async def test_ourocg_multiple(ourocg: OurOcg):
    cards = await ourocg.list_from_query("宝札", limit=6)
    assert cards
    assert len(cards) == 6
    assert all(CardType.Spell in card.types for card in cards)
    assert ourocg.current_query.query == "宝札"
    assert ourocg.current_query.current == 6
    assert ourocg.current_query.total_page > 1

@pytest.mark.asyncio
async def test_ourocg_ids(ourocg: OurOcg):
    ids = await ourocg.list_ids_from_query("迪安·凯特", limit=20)
    assert ids
    assert len(ids) < 20
    assert all(isinstance(card, OurOcgIDCard) for card in ids)
    assert any(card.id == 0 for card in ids)  # RD
    



from typing import TypeVar
from enum import IntFlag
from functools import partial

from ygoutil.card import Card
from ygoutil.card.unit import AliasIDUnit, LimitUnit, SetUnit, LinkUnit, CategoryUnit, PTextUnit
from ygoutil.card.constant import CardType, CardRace, CardAttribute, CardCategory, LinkMark, CardOT, CardLF

def deal_atk_def(val: int):
    return val if val >= 0 else "?"

def deal_level(val: int):
    return val & 0b1111

def deal_p_mark(val: int):
    pl = (val & 0xF000000) >> 24
    pr = (val & 0x00F0000) >> 16
    return pl, pr

FlagVar = TypeVar("FlagVar", bound=IntFlag)

def bits2set(bits: int, flags: type[FlagVar]) -> set[FlagVar]:
    return {x for x in flags if x & bits != 0}

def bits2item(bits: int, flags: type[FlagVar]) -> FlagVar:
    r = flags(0)
    any((r := x) for x in flags if x & bits != 0)
    return r

bits2CardTypes = partial(bits2set, flags=CardType)
bits2Race = partial(bits2item, flags=CardRace)
bits2Attribute = partial(bits2item, flags=CardAttribute)
bits2LinkMark = partial(bits2item, flags=LinkMark)
bits2Category = partial(bits2item, flags=CardCategory)

def card_from_types(types: int):
    return Card.from_types(*bits2CardTypes(types))

def card_from_cdb_text(card: Card, name: str, text: str) -> Card:
    card.name = name
    card.text = text
    return card

def card_from_cdb_data(card: Card, id: int, ot: int, alias: int, sets: int, types: int, 
                      attack: int, defence: int, level: int, race: int, attribute: int, category: int,
                      set_map: dict[int, str] = None, lf_map: dict[int, int] = None) -> Card:
    if alias:
        id, alias = alias, id  # alias 里面才是真的卡号，id 里面是替身卡号
        card.ids = AliasIDUnit(card)
        card.ids.alias_id = alias
    card.id = id
    card._limit_unit = LimitUnit(card)  # 有 LimitUnit
    card.limits.ot = CardOT(ot)
    if lf_map:
        card.limits.limit = CardLF(lf_map.get(card.id, CardLF.无限制))
    if set_map:
        card._set_unit = SetUnit(card)  # 传入 set_map 的话，即使没有字段，也保证 SetUnit 存在
        while sets != 0:
            if set_name := set_map.get(sets & 0xFFFF):
                card.sets.add(set_name)
            sets = sets >> 16
    if card.is_monster:
        monster = card.monster
        monster.attack = deal_atk_def(attack)
        monster.level = deal_level(level)
        if isinstance(monster.levels, LinkUnit):
            monster.levels.marks = LinkMark(defence)  # 连接怪用防御记录连接标记
        else:
            monster.defence = deal_atk_def(defence)
        if monster._pmark_unit:
            monster._pmark_unit.mark = deal_p_mark(level)
            card.texts = PTextUnit(card)  # 灵摆怪兽试用 PTextUnit
        monster.race = CardRace(race)
        monster.attribute = CardAttribute(attribute)
    card._category_unit = CategoryUnit(card)  # 有 CategoryUnit
    card.categories = CardCategory(category)
    return card

def card_from_cdb(name: str, text: str, 
                      id: int, ot: int, alias: int, sets: int, types: int, 
                      attack: int, defence: int, level: int, race: int, attribute: int, category: int,
                      set_map: dict[int, str] = None, lf_map: dict[int, int] = None) -> Card:
    card = card_from_types(types)
    card = card_from_cdb_data(card, id, ot, alias, sets, types, 
                              attack, defence, level, race, attribute, category, set_map, lf_map)
    card = card_from_cdb_text(card, name, text)
    return card



from dataclasses import dataclass

@dataclass
class IDCard:
    """ 卡片基础标识 """
    id: int | str = 0
    name: str = ""

    def __bool__(self):
        return bool(self.id or self.name)

@dataclass
class CnJpEnIDCard(IDCard):
    """ 多语言卡片基础标识 """
    jp_name: str = ""
    en_name: str = ""

    def __bool__(self):
        return super().__bool__() or bool(self.jp_name or self.en_name)
